import os
import socket
import datetime
import platform

import consoleiotools as cit
import consolecmdtools as cct


class Trackfile:
    def __init__(self, trackfile_dir: str = os.getcwd(), prefix: str = "FileTrack-", format: str = "json", group_by: str = ""):
        """Initialize Trackfile object.

        Args:
            trackfile_dir (str): The directory path of the trackfile.
            prefix (str): The prefix of the trackfile.
            format (str): The output format. Options: "json", "toml".
            group_by (str): Group by. Default is "" meaning no group. Options: "host", "os", "".
        """
        self.prefix = prefix
        self.trackfile_dir = trackfile_dir
        self.format = format
        if self.format.upper() == "TOML":
            import tomlkit  # lazyload
            self.suffix = ".toml"
            self.formatter = tomlkit
        elif self.format.upper() == "JSON":
            import json  # lazyload
            self.suffix = ".json"
            self.formatter = json
        else:
            raise Exception(f"Output format `{self.format}` does not support")
        self.group_by = group_by
        self.trackings = {}

    def __str__(self) -> str:
        return self.path

    @property
    def files(self) -> list:
        def filename_filter(path: str) -> bool:
            filename = cct.get_path(path).basename
            if filename.startswith(self.prefix) and filename.endswith(self.suffix):
                if self.group_by and (self.group not in filename):
                    return False
                return True
            return False

        trackfile_list = cct.get_paths(self.trackfile_dir, filter=filename_filter)
        return sorted(trackfile_list)

    @property
    def latest(self) -> str:
        if not self.files:
            return ""
        return self.files[-1]

    @property
    def group(self) -> str:
        if self.group_by == "host":
            return self.hostname
        elif self.group_by == "os":
            return platform.system()
        elif self.group_by == "":
            return ""
        cit.err(f"Unsupported group_by: {self.group_by}")
        exit(1)

    @property
    def filename(self):
        now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        return f"{self.prefix}{now}{'-' if self.group else ''}{self.group}{self.suffix}"

    @property
    def path(self):
        return cct.get_path(os.path.join(self.trackfile_dir, self.filename))

    def compare_with(self, trackfile: "Trackfile") -> tuple[list, list]:
        trackings_1 = set(self.trackings.items())
        trackings_2 = set(trackfile.trackings.items())
        return [filename for filename, filehash in trackings_1 - trackings_2], [filename for filename, filehash in trackings_2 - trackings_1]

    @property
    def hostname(self) -> str:
        """Get hostname and check if hostname is new."""
        host = socket.gethostname().replace("-", "").replace(".", "")  # default hostname
        return host

    @cit.as_session("Cleanup Outdated TrackFiles")
    def cleanup_outdated_trackfiles(self):
        if len(self.files) > 1:
            old_trackfiles = self.files[:-1]  # exclude the latest one
            cit.ask(f"Cleanup {len(old_trackfiles)} old TrackFiles?")
            for trackfile in old_trackfiles:
                cit.echo(cct.get_path(trackfile).basename, pre="*")
            if cit.get_choice(["Yes", "No"]) == "Yes":
                for filepath in old_trackfiles:
                    os.remove(filepath)
                cit.info("Cleanup done")
            else:
                cit.warn("Cleanup canceled")

    @cit.as_session("Saving TrackFile")
    def to_file(self):
        with open(self.path, "w", encoding="UTF8") as f:
            options = {}
            if self.format == "JSON":
                options = {"indent": 4, "ensure_ascii": False}
            f.write(self.formatter.dumps(self.trackings, **options))
            cit.info(f"New TrackFile created: [u]{self.path.basename}[/]")

    @cit.as_session("Loading Old TrackFile")
    def from_file(self, path: str) -> bool:
        """Parse a TrackFile and load trackings into instance.

        Args:
            path (str): Path to the TrackFile.

        Returns:
            bool: True if successful.
        """
        if not path:
            return False
        path = cct.get_path(path)
        if not path.is_file:
            cit.warn("No TrackFile loaded.")
            return False
        cit.info(f"Parsing TrackFile `{path.basename}`")
        with open(path, encoding="UTF8") as fl:
            trackings = self.formatter.loads(fl.read())
            cit.info(f"{len(trackings)} entries loaded")
        self.trackings = trackings
        return True

    def target_files(self, target_dir: str = os.getcwd(), exts: list = []) -> list:
        """Get target files in the target directory.

        Args:
            target_dir (str): Target directory to scan.
            exts (list[str]): Accepted file extensions. Ex. ["mp3", "m4a"]. Default is [] meaning all files.

        Returns:
            list: List of target file paths.
        """
        paths = []
        if not exts:
            paths += cct.get_paths(target_dir, filter=os.path.isfile)
        else:
            for ext in exts:
                target_file_pattern = f".{ext}"
                cit.info(f"Target file pattern: {target_file_pattern}")
                paths += cct.get_paths(target_dir, filter=lambda path: path.name.endswith(target_file_pattern))
        return paths

    @cit.as_session("Generating New TrackFile")
    def generate(self, target_dir: str = os.getcwd(), exts: list = [], hash_mode: str = "HASH"):
        """Generate file tracking information.

        Args:
            target_dir (str): Target directory to scan.
            exts (list[str]): Accepted file extensions. Ex. ["mp3", "m4a"]. Default is [] meaning all files.
            hash_mode (str): "HASH", "CRC32", "MD5", "NAME", "PATH", "MTIME".

        Returns:
            dict: {filename: filehash}
        """
        def find_duplicates(paths: list) -> list:
            """Find duplicate filename files in the list of paths.

            Returns:
                list: List of duplicate filepaths.
            """
            duplicate_files = set()
            files = {}
            for filepath in paths:
                filepath = cct.get_path(filepath)
                if path := files.get(filepath.basename):
                    duplicate_files.add(path)
                    duplicate_files.add(filepath)
                else:
                    files[filepath.basename] = filepath
            return list(duplicate_files)

        paths = self.target_files(target_dir, exts)
        cit.info(f"Target files: {len(paths)}")
        if paths:
            duplicate_files = find_duplicates(paths)
            cit.info(f"Duplicate files: {len(duplicate_files)}")
            for filepath in cit.track(paths, "Hashing...", unit="files"):
                filepath = cct.get_path(filepath)
                match hash_mode.lower():  # >=3.10
                    case "hash":
                        filehash = abs(hash(filepath))
                    case "crc32":
                        filehash = cct.crc32(filepath)
                    case "md5":
                        filehash = cct.md5(filepath)
                    case "mtime":
                        filehash = int(os.path.getmtime(filepath))
                    case "name":
                        filehash = filepath.basename
                    case "path":
                        filehash = filepath.abs
                    case _:
                        cit.err(f"Unsupported hash mode: {hash_mode}")
                        exit(1)
                filehash = str(filehash)
                if oldhash := self.trackings.get(filepath.basename):
                    self.trackings[filepath.basename] = ",".join([oldhash, filehash])
                else:
                    self.trackings[filepath.basename] = filehash



import os
import tomllib

import fuzzyfinder
import consolecmdtools as cct
import consoleiotools as cit

from . import Trackfile


__version__ = '0.1.0'


def load_configs(config_path: str, target_dir: str, target_exts: list[str], trackfile_dir: str, trackfile_format: str, hash_mode: str, group_by: str) -> dict:
    """Get configurations from the config file.

    Args:
        config_path (str, optional): The path or the file name of the config file. Defaults to "filetrack.toml".

    Returns:
        dict: The configurations. If the file does not exist, return an empty dict.
    """
    configs = {}
    if config_path:
        if os.path.isfile(config_path):
            with open(config_path, "rb") as fl:
                configs = tomllib.load(fl)
        else:
            cit.warn(f"Config file not found: [u]{config_path}[/]")
    # [target] defaults
    if not configs.get("target"):
        configs['target'] = {}
    if not configs["target"].get("dir"):
        configs['target']['dir'] = target_dir
    if not os.path.isabs(configs['target']['dir']):
        configs['target']['dir'] = os.path.join(cct.get_path(config_path).parent, configs['target']['dir'])
    configs['target']['dir'] = cct.get_path(configs['target']['dir'])
    if not configs['target'].get("exts"):
        configs['target']['exts'] = target_exts
    # [trackfile] defaults
    if not configs.get("trackfile"):
        configs['trackfile'] = {}
    configs['trackfile']['dir'] = configs['trackfile'].get("dir") or trackfile_dir
    if not os.path.isabs(configs['trackfile']['dir']):
        configs['trackfile']['dir'] = os.path.join(cct.get_path(config_path).parent, configs['trackfile']['dir'])
    configs['trackfile']['dir'] = cct.get_path(configs['trackfile']['dir'])
    if not configs['trackfile'].get("format"):
        configs['trackfile']['format'] = trackfile_format
    if not configs['trackfile'].get("hash_mode"):
        configs['trackfile']['hash_mode'] = hash_mode
    if not configs['trackfile'].get("group_by"):
        configs['trackfile']['group_by'] = group_by
    if not configs['trackfile'].get("prefix"):
        configs['trackfile']['prefix'] = "FileTrack-"
    return configs


def diffs(old_ft: Trackfile.Trackfile, new_ft: Trackfile.Trackfile) -> bool:
    """Check if there are any differences between the old and new trackfiles.

    Args:
        old_ft (Filetrack.Trackfile): The old trackfile.
        new_ft (Filetrack.Trackfile): The new trackfile.

    Returns:
        bool: True if there are differences, False otherwise.
    """
    pres = {
        "delete": "[red]-[/]",
        "add": "[green]+[/]",
        "move_from": "[blue]-[/]",
        "move_to": "[blue]+[/]",
    }
    if not old_ft.trackings and not new_ft.trackings:
        return False
    entries_deleted, entries_added = old_ft.compare_with(new_ft)
    if not entries_deleted and not entries_added:
        cit.info("No changes")
        return False
    cit.info("Changes since last time: ✍️")
    for filename in entries_deleted:
        if entries_added:
            fuzzy = list(fuzzyfinder.fuzzyfinder(filename, entries_added))
            if len(fuzzy) > 0:
                cit.echo(f"{pres['move_from']} {filename}")
                cit.echo(f"{pres['move_to']} {fuzzy[0]}")
                entries_added.remove(fuzzy[0])
            else:
                cit.echo(f"{pres['delete']} {filename}")
    for filename in entries_added:
        cit.echo(f"{pres['add']} {filename}")
    return True


def run_filetrack(config_path: str = "filetrack.toml", target_dir: str = ".", target_exts: list[str] = [], trackfile_dir: str = ".", trackfile_format: str = "json", hash_mode: str = "HASH", group_by: str = ""):
    """Run filetrack

    Args:
        config_path (str, optional): The path or the file name of the config file. Defaults to "filetrack.toml".
        target_dir (str): The directory path of the files to be tracked. Default is config file's parent directory.
        target_exts (list[str], optional): The target extensions. Defaults to TARGET_EXTS.
        trackfile_dir (str): The directory path of the trackfile. Default is config file's parent directory.
        trackfile_format (str, optional): The output format. Defaults to "json". Options: "json", "toml".
        hash_mode (str, optional): The hash mode. Defaults to "HASH". Options: "HASH", "CRC32", "MD5", "NAME", "PATH", "MTIME".
        group_by (str, optional): Group by. Defaults to "". Options: "host", "os", "".
    """
    cit.rule("▶ [yellow]Run Filetrack[/]")
    configs = load_configs(config_path, target_dir, target_exts, trackfile_dir, trackfile_format, hash_mode, group_by)
    old_ft = Trackfile.Trackfile(trackfile_dir=configs['trackfile']['dir'], prefix=configs['trackfile']['prefix'], format=configs['trackfile']['format'], group_by=configs['trackfile']['group_by'])
    new_ft = Trackfile.Trackfile(trackfile_dir=configs['trackfile']['dir'], prefix=configs['trackfile']['prefix'], format=configs['trackfile']['format'], group_by=configs['trackfile']['group_by'])
    cit.info(f"Version: {__version__}")
    cit.info(f"Config File: [u]{config_path}[/]")
    cit.info(f"Target Dir: 📂 [u]{configs['target']['dir']}[/]")
    cit.info(f"Target Extensions: {configs['target']['exts']}")
    cit.info(f"Trackfile Format: {configs['trackfile']['format']}")
    cit.info(f"Trackfile Dir: 📂 [u]{configs['trackfile']['dir']}[/]")
    cit.info(f"Trackfile Prefix: `{configs['trackfile']['prefix']}`")
    cit.info(f"Hash Mode: {configs['trackfile']['hash_mode']}")
    if configs['trackfile']['group_by']:
        cit.info(f"Group by {configs['trackfile']['group_by']}: `{new_ft.group}`")
    # fill trackings
    old_ft.from_file(old_ft.latest)
    new_ft.generate(configs['target']['dir'], configs['target']['exts'], configs['trackfile']['hash_mode'])
    if not old_ft.trackings or diffs(old_ft, new_ft):
        cit.info("[Done]")
        new_ft.to_file()
        new_ft.cleanup_outdated_trackfiles()



import argparse

import consoleiotools as cit

import filetrack


README_URL = "https://github.com/kyan001/PyFileTrack/blob/main/README.md"


def main(assigned_args: list | None = None):
    parser = argparse.ArgumentParser(prog="filetrack", description="Tracking file changes.", epilog=f"Checkout README for more details: {README_URL}")
    parser.add_argument("-v", "--version", action="version", version=filetrack.__version__)
    parser.add_argument("-c", "--config", dest="config", default="filetrack.toml", help="The path or the file name of the config file.")
    args = parser.parse_args(assigned_args)
    filetrack.run_filetrack(args.config)


if __name__ == "__main__":
    main()
    cit.pause()



import sys
import os
import io
import unittest
from unittest.mock import patch

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from filetrack import command_line  # noqa: linter (pycodestyle) should not lint this line.


class test_filetrack(unittest.TestCase):
    """command-line filetrack unittest"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_no_args(self):
        with patch("sys.stdout", new=io.StringIO()) as fake_out:
            command_line.main([])
            self.assertIn("filetrack.toml", fake_out.getvalue())

    def test_config(self):
        with patch("sys.stdout", new=io.StringIO()) as fake_out:
            command_line.main(["-c", "filetrack.toml"])
            self.assertIn("filetrack.toml", fake_out.getvalue())
            command_line.main(["--config", "filetrack.toml"])
            self.assertIn("filetrack.toml", fake_out.getvalue())


if __name__ == "__main__":
    unittest.main(verbosity=2, exit=False)



import sys
import os.path
import unittest

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import filetrack  # noqa: linter (pycodestyle) should not lint this line.


class test_filetrack(unittest.TestCase):
    """filetrack unittest"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_version(self):
        self.assertIsInstance(filetrack.__version__, str)


if __name__ == "__main__":
    unittest.main(verbosity=2, exit=False)
